@@ -1,4 +1,4 @@
-/* Copyright 1996-1998,2000-2007,2009,2011-2015,2018,2021
+/* Copyright 1996-1998,2000-2007,2009,2011-2015,2018,2021,2022
      Free Software Foundation, Inc.
 
    This file is part of Guile.
@@ -213,15 +213,18 @@ SCM_DEFINE (scm_inet_makeaddr, "inet-makeaddr", 2, 0, 0,
 static SCM
 scm_from_ipv6 (const uint8_t *src)
 {
-  SCM result = scm_i_mkbig ();
-  mpz_import (SCM_I_BIG_MPZ (result),
+  mpz_t z;
+  mpz_init (z);
+  mpz_import (z,
               1,  /* chunk */
               1,  /* big-endian chunk ordering */
               16, /* chunks are 16 bytes long */
               1,  /* big-endian byte ordering */
               0,  /* "nails" -- leading unused bits per chunk */
               src);
-  return scm_i_normbig (result);
+  SCM ret = scm_from_mpz (z);
+  mpz_clear (z);
+  return ret;
 }
 
 /* convert a host ordered SCM integer to a 128 bit IPv6 address in
@@ -252,10 +255,16 @@ scm_to_ipv6 (uint8_t dst[16], SCM src)
   else if (SCM_BIGP (src))
     {
       size_t count;
+      mpz_t z;
       
-      if ((mpz_sgn (SCM_I_BIG_MPZ (src)) < 0)
-	  || mpz_sizeinbase (SCM_I_BIG_MPZ (src), 2) > 128)
-	scm_out_of_range (NULL, src);
+      mpz_init (z);
+      scm_to_mpz (src, z);
+
+      if (mpz_sgn (z) < 0 || mpz_sizeinbase (z, 2) > 128)
+        {
+          mpz_clear (z);
+          scm_out_of_range (NULL, src);
+        }
       
       memset (dst, 0, 16);
       mpz_export (dst,
@@ -264,8 +273,8 @@ scm_to_ipv6 (uint8_t dst[16], SCM src)
                   16, /* chunks are 16 bytes long */
                   1, /* big-endian byte ordering */
                   0, /* "nails" -- leading unused bits per chunk */
-                  SCM_I_BIG_MPZ (src));
-      scm_remember_upto_here_1 (src);
+                  z);
+      mpz_clear (z);
     }
   else
     scm_wrong_type_arg_msg ("scm_to_ipv6", 0, src, "integer");
@@ -1491,21 +1500,24 @@ SCM_DEFINE (scm_recvfrom, "recvfrom!", 2, 3, 0,
 
   SCM_VALIDATE_BYTEVECTOR (1, buf);
 
-  if (SCM_UNBNDP (start))
-    offset = 0;
-  else
-    offset = scm_to_size_t (start);
-
   if (SCM_UNBNDP (end))
     cend = SCM_BYTEVECTOR_LENGTH (buf);
   else
     {
       cend = scm_to_size_t (end);
-      if (SCM_UNLIKELY (cend >= SCM_BYTEVECTOR_LENGTH (buf)
-                        || cend < offset))
+      if (SCM_UNLIKELY (cend > SCM_BYTEVECTOR_LENGTH (buf)))
         scm_out_of_range (FUNC_NAME, end);
     }
 
+  if (SCM_UNBNDP (start))
+    offset = 0;
+  else
+    {
+      offset = scm_to_size_t (start);
+      if (SCM_UNLIKELY (cend < offset))
+        scm_out_of_range (FUNC_NAME, start);
+    }
+
   SCM_SYSCALL (rv = recvfrom (fd,
                               SCM_BYTEVECTOR_CONTENTS (buf) + offset,
                               cend - offset, flg,
