@@ -1,6 +1,6 @@
 ;;; Continuation-passing style (CPS) intermediate language (IL)
 
-;; Copyright (C) 2013, 2014, 2015, 2017, 2018 Free Software Foundation, Inc.
+;; Copyright (C) 2013-2021 Free Software Foundation, Inc.
 
 ;;;; This library is free software; you can redistribute it and/or
 ;;;; modify it under the terms of the GNU Lesser General Public
@@ -80,15 +80,19 @@ sites."
                                      (causes-effect? fx &allocation))
                                 (values (intset-add! known k) unknown)
                                 (values known (intset-add! unknown k)))))
-                         (($ $kargs _ _ (or ($ $branch) ($ $prompt) ($ $throw)))
-                          ;; Branches and prompts pass no values to
-                          ;; their continuations, and throw terms don't
-                          ;; continue at all.
+                         (($ $kargs _ _ (or ($ $branch) ($ $switch)
+                                            ($ $prompt) ($ $throw)))
+                          ;; Branches, switches, and prompts pass no
+                          ;; values to their continuations, and throw
+                          ;; terms don't continue at all.
                           (values known unknown))
                          (($ $kreceive arity kargs)
                           (values known (intset-add! unknown kargs)))
-                         (($ $kfun src meta self tail clause)
-                          (values known unknown))
+                         (($ $kfun src meta self tail entry)
+                          (values known
+                                  (if entry
+                                      (intset-add! unknown entry)
+                                      unknown)))
                          (($ $kclause arity body alt)
                           (values known (intset-add! unknown body)))
                          (($ $ktail)
@@ -153,7 +157,9 @@ sites."
          (values live-labels (adjoin-vars args (adjoin-var proc live-vars))))
         (($ $callk kfun proc args)
          (values (intset-add live-labels kfun)
-                 (adjoin-vars args (adjoin-var proc live-vars))))
+                 (adjoin-vars args (if proc
+                                       (adjoin-var proc live-vars)
+                                       live-vars))))
         (($ $primcall name param args)
          (values live-labels (adjoin-vars args live-vars)))
         (($ $values args)
@@ -202,7 +208,8 @@ sites."
         ;; Still dead.
         (values live-labels live-vars))))
 
-    (define (visit-branch label kf kt args live-labels live-vars)
+    ;; Note, this is for $branch or $switch.
+    (define (visit-branch label kf kt* args live-labels live-vars)
       (define (next-live-term k)
         ;; FIXME: For a chain of dead branches, this is quadratic.
         (let lp ((seen empty-intset) (k k))
@@ -214,12 +221,23 @@ sites."
               (($ $kargs _ _ ($ $continue k*))
                (lp (intset-add seen k) k*))
               (_ k))))))
+      (define (distinct-continuations?)
+        (let ((kf' (next-live-term kf)))
+          (let lp ((kt* kt*))
+            (match kt*
+              (() #f)
+              ((kt . kt*)
+               (cond
+                ((or (eqv? kf kt)
+                     (eqv? kf' (next-live-term kt)))
+                 (lp kt*))
+                (else #t)))))))
       (cond
        ((intset-ref live-labels label)
         ;; Branch live already.
         (values live-labels (adjoin-vars args live-vars)))
        ((or (causes-effect? (intmap-ref effects label) &type-check)
-            (not (eqv? (next-live-term kf) (next-live-term kt))))
+            (distinct-continuations?))
         ;; The branch is live if its continuations are not the same, or
         ;; if the branch itself causes type checks.
         (values (intset-add live-labels label)
@@ -236,7 +254,9 @@ sites."
            (($ $kargs _ _ ($ $continue k src exp))
             (visit-exp label k exp live-labels live-vars))
            (($ $kargs _ _ ($ $branch kf kt src op param args))
-            (visit-branch label kf kt args live-labels live-vars))
+            (visit-branch label kf (list kt) args live-labels live-vars))
+           (($ $kargs _ _ ($ $switch kf kt* src arg))
+            (visit-branch label kf kt* (list arg) live-labels live-vars))
            (($ $kargs _ _ ($ $prompt k kh src escape? tag))
             ;; Prompts need special elision passes that would contify
             ;; aborts and remove corresponding "unwind" primcalls.
@@ -250,8 +270,11 @@ sites."
             (values live-labels live-vars))
            (($ $kclause arity kargs kalt)
             (values live-labels (adjoin-vars (cont-defs kargs) live-vars)))
-           (($ $kfun src meta self)
-            (values live-labels (adjoin-var self live-vars)))
+           (($ $kfun src meta self tail entry)
+            (values live-labels
+                    (adjoin-vars
+                     (or (and entry (cont-defs entry)) '())
+                     (if self (adjoin-var self live-vars) live-vars))))
            (($ $ktail)
             (values live-labels live-vars))))
        conts label live-labels live-vars))
@@ -354,6 +377,11 @@ sites."
            ;; Dead branches continue to the same continuation
            ;; (eventually).
            (values cps (build-term ($continue kf src ($values ()))))))
+      (($ $switch kf kt* src arg)
+       ;; Same as in $branch case.
+       (if (label-live? label)
+           (values cps term)
+           (values cps (build-term ($continue kf src ($values ()))))))
       (($ $prompt)
        (values cps term))
       (($ $throw)
